#define	config_file	"/etc/glftpd.conf"
#define KEY	0x0000DEAD   // Default KEY used by DAEMON

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/time.h>
#include <time.h>
#include <errno.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include <syslog.h>
#include <unistd.h>
#include <sys/param.h>

struct ONLINE
{
  char   tagline[64];
  char   username[24];
  char   status[256];
  char   host[256];
  char   currentdir[256];
  long   groupid;
  time_t login_time;
  struct timeval tstart;   
  unsigned long bytes_xfer; 
  pid_t  procid;
};

static struct ONLINE *online;
static int  dl = 0;
static int  ul = 0;
static int  max_users = 0;
static int  shmid;
static void quit();
static double calc_time();




static char *
trim( char *str )
{
	char *ibuf;
	char *obuf;

   if ( str )
   {
   	for ( ibuf = obuf = str; *ibuf; )
      {
        	while ( *ibuf && ( isspace ( *ibuf ) ) )
         	ibuf++;
        	if ( *ibuf && ( obuf != str ) )
            *( obuf++ ) = ' ';
        	while ( *ibuf && ( !isspace ( *ibuf ) ) )
            *( obuf++ ) = *( ibuf++ );
      }
      *obuf = '\0';
   }
   return( str );
}
				 
static void
load_sysconfig( void )
{
   FILE *configfile;
   char lvalue[ 64 ];
   char rvalue[ MAXPATHLEN ];
   int  x, y;
char work_buff[ MAXPATHLEN ];

 sprintf( work_buff, "%s", config_file);

   if ( ( configfile = fopen( work_buff, "r" ) ) == NULL )
	{
		fprintf( stderr, "bad or missing 'CONFIG FILE'\n" );
		exit (1);
	}

   while ( 1 )
   {
      if ( fgets( work_buff, sizeof( work_buff ), configfile ) == NULL )
      {
         fclose( configfile );
         return;
      }

      /* Clip out comments */
      for ( x = 0; x < strlen( work_buff ); x++ )
         if ( work_buff[ x ] == '#' )
            work_buff[ x ] = '\0';
				
		/* Trim */
		(void)trim( work_buff );
      
      /* Clear out old values */
      memset( lvalue, '\0', sizeof( lvalue ) );
      memset( rvalue, '\0', sizeof( rvalue ) );
													 
		/* The file allows spaces embedded, but clears leading & trailing 
			spaces */
		
      /* Parse lvalue */
      y = 0;
      for ( x = 0; x < strlen( work_buff ) && work_buff[ x ] != ' '; x++ )
         if ( isprint( work_buff[ x ] ) )
            lvalue[ y++ ] = work_buff[ x ];

      /* Parse rvalue */
      y = 0; x++;
      for ( ; x < strlen( work_buff ); x++ )
         if ( isprint( work_buff[ x ] ) )
            rvalue[ y++ ] = work_buff[ x ];
      
      if ( strcasecmp( lvalue, "max_users" ) == 0 )
	   	max_users = atol( rvalue );
   }
   return;
}


static char hmsbuf[50];			
void
pid_hms_format( time_t testtime )
{
   time_t timenow = time( NULL );
   time_t difftime;
   int    days = 0;
   int    hours = 0;
   int    minutes = 0;
	int    seconds = 0;

   
   difftime = timenow - testtime;
   
   while ( difftime >= (time_t)3600 )
   {
      difftime -= (time_t)3600;
      hours++;
   }          
   
   while ( difftime >= (time_t)60 )
   {
      difftime -= (time_t)60;
      minutes++;
   }
   
   if ( hours != 0 )
      sprintf( hmsbuf, "%02d:%02d:%02d", hours, minutes, difftime );
   else
      sprintf( hmsbuf, "   %02d:%02d", minutes, difftime );
}


static int
checkusers(void)
{
  register int i, found = 0;
  char statbuf[ 500 ];
  char pidbuf[ 10 ];
															
			
	for (i=0; i<max_users; i++) 
	{
		if (online[i].procid == 0)
			continue;
		sprintf( pidbuf, "%u", online[i].procid );
		pid_hms_format( online[i].tstart.tv_sec );  
		
      		if (!strncasecmp (online[i].status, "STOR", 4) && online[i].bytes_xfer != 0)
		{
				sprintf( statbuf, "Up: %-17.17s at %.1fK/sec", online[i].status+5, calc_time(i) );
				ul++;
		}
	      /* Downloading */
	        else if (!strncasecmp (online[i].status, "RETR", 4) && online[i].bytes_xfer != 0)
		{
				sprintf( statbuf, "Dn: %-17.17s at %.1fK/sec", online[i].status+5, calc_time(i) );
				dl++;
		}
	      /* Idling */
   	   	else if (online[i].tstart.tv_sec - time(NULL) > 5)
		{
				sprintf( statbuf, "Idle: %-8.8s", hmsbuf );
		}
   	   /* Doing something else... */
	        else
		{
				sprintf( statbuf, "\"%s\"", trim(online[i].status) );
		}
		pid_hms_format( online[i].login_time); 
		fprintf( stdout, "| %-8.8s | %-5.5s | %-8.8s | %-44.44s |\n", online[i].username, pidbuf, hmsbuf, statbuf);
	  	found++;
	}
  return (found);
}


                                   
int
main( int argc, char **argv )
{
	int numusers;

	load_sysconfig();	
        if ((shmid = shmget( (key_t)KEY, (int)(sizeof(struct ONLINE) * max_users), 0)) == -1)
        {
                fprintf(stdout, "No Users Currently On Site!\n");
                exit(1);
        }
        if ((online = (struct ONLINE *)shmat( shmid, NULL, SHM_RDONLY)) == (struct ONLINE *)-1)
        {
                fprintf(stdout, "Error: (SHMAT) Failed!\n");
                exit(1);
        }
							     
	fprintf(stdout,".----------------------------------------------------------------------------.\n" );
	fprintf(stdout,"| Username | PID   | OnLine   | Action                                       |\n" );
	fprintf(stdout,"|----------+-------+----------+----------------------------------------------|\n" );
	numusers = checkusers();
	fprintf(stdout,"`----------------------------------------------------------------------------'\n" );   					  
   
   quit( 0 );
}


static double
calc_time(int pid)
{
        struct timeval tstop;
        double delta, rate;

        if (online[pid].bytes_xfer < 1)
                return 0;
        gettimeofday(&tstop, (struct timezone *)0 );
        delta = ((tstop.tv_sec*10.)+(tstop.tv_usec/100000.)) -
                ((online[pid].tstart.tv_sec*10.)+(online[pid].tstart.tv_usec/100000.));
        delta = delta/10.;
        rate = ((online[pid].bytes_xfer / 1024.0) / (delta));
        if (!rate) rate++;
        return (double)(rate);
}



static void
quit(int exit_status)
{
        struct shmid_ds ipcbuf;

        shmctl( shmid, IPC_STAT, &ipcbuf);
        if (ipcbuf.shm_nattch <= 1) {
               shmctl( shmid, IPC_RMID, 0);
        }
        shmdt(0);
        exit(exit_status);
}

